from dovado_rtl.generators.generator import Generator
from nacolla.models import ImmutableModel
from dovado_rtl.explorers.utilities.design_points import DesignPoint
import subprocess
from pathlib import Path


class GeneratedProject(ImmutableModel):
    command: str
    project_root: str
    target_file: str
    target_module: str


class BashGenerator(Generator):
    def generate(self, design_point: DesignPoint) -> DesignPoint:
        if design_point.extra is None or design_point.extra["generate"] is None:
            raise ValueError(
                "Please add an extra.generate section with the command to run and the new project root"
            )
        generated_project = GeneratedProject(**design_point.extra["generate"])

        bash_command = subprocess.Popen(
            generated_project.command,
            shell=True,
            stdout=subprocess.PIPE,
            cwd=design_point.work_directory,
        )
        bash_command.wait()

        return DesignPoint(
            **(
                dict(design_point)
                | {
                    "project_root": Path(
                        design_point.work_directory, generated_project.project_root
                    ),
                    "target_file": generated_project.target_file,
                    "target_module": generated_project.target_module,
                }
            )
        )



#!/usr/bin/env python3
from typing import Sequence, cast

from antlr4 import TokenStream
from dovado_rtl.parsers.utilities.antlr.antlr_parsed import AntlrParsed
from dovado_rtl.parsers.utilities.antlr.hdl.hdl_antlr_module import HdlAntlrModule


class HdlAntlrParsed(AntlrParsed):
    def __init__(
        self, token_stream: TokenStream, modules: Sequence[HdlAntlrModule]
    ) -> None:
        super().__init__(token_stream, modules)
        self._modules = modules

    @property
    def modules(self) -> Sequence[HdlAntlrModule]:
        return self._modules

    def get_module(self, name: str) -> HdlAntlrModule:
        return cast(HdlAntlrModule, super().get_module(name))



from pathlib import Path
from typing import Any, Callable, Iterable, Literal, Optional, Union
from movado import approximate
from pymoo.core.problem import ElementwiseProblem, Problem
from pymoo.core.repair import Repair
from pymoo.core.algorithm import Algorithm
from pymoo.core.result import Result
from pymoo.core.termination import Termination
from pymoo.factory import get_crossover, get_mutation, get_sampling, get_termination
from pymoo.optimize import minimize
from dovado_rtl.explorers.utilities.design_points import (
    DesignPoint,
    EvaluatedDesignPoint,
)
from dovado_rtl.explorers.explorer import EndExploration, Explorer
import numpy as np
from dovado_rtl.explorers.utilities.spaces import SOURCE_PATH
from dovado_rtl.explorers.utilities.tasks import (
    AutomaticExplorationProject,
    ParsedProject,
)

from threading import Lock

from dovado_rtl.parsers.utilities.parsed import MODULE_NAME_STR, PARAMETER_NAME_STR

from queue import SimpleQueue

import importlib
from threading import Thread


class EndResult(EndExploration):
    result: Result


class AutomaticExplorer(Explorer):
    __pymoo_package_structure = "pymoo.algorithms.moo"
    __seed = 0
    __parameter_file_name = "parameters"
    __metrics_file_name = "metrics"

    def __init__(self) -> None:
        super().__init__()
        self._approximation_to_initialize = True
        self._parameters_structure: dict[
            SOURCE_PATH, dict[MODULE_NAME_STR, list[PARAMETER_NAME_STR]]
        ] = {}
        self._parsed_project: Optional[ParsedProject] = None

        self._evaluated_points: list[tuple[list[int], EvaluatedDesignPoint]] = []
        self._points_under_evaluation: list[tuple[list[int], Lock]] = []
        self._points_queue: SimpleQueue[
            Union[list[Union[list[int], bool]], EndResult]
        ] = SimpleQueue()
        self._metrics: list[str] = []
        self._task: Optional[AutomaticExplorationProject] = None

    def update(
        self, evaluated_design_point: Union[EvaluatedDesignPoint, EndExploration]
    ) -> Union[DesignPoint, EndExploration]:
        if isinstance(evaluated_design_point, EndExploration):
            return EndExploration()

        self._evaluated_points.append(
            (evaluated_design_point.vectorized(), evaluated_design_point)
        )

        self._release(evaluated_design_point)

        point = self._get_next_design_point()

        if isinstance(point, EndResult):
            self._save_results(point.result)
            return EndExploration()

        if self._task is None:
            raise ValueError("Please save the task in explore, found it None")

        self._set_point(point, self._task)

        if self._parsed_project is None:
            raise ValueError(
                "Found None parsed_project in automatic explorer, please save it in the explore method"
            )

        return DesignPoint.make_design_point_from_vector(
            vector=point,
            parameters_structure=self._parameters_structure,
            parsed_project=self._parsed_project,
        )

    def explore(self, task: AutomaticExplorationProject) -> DesignPoint:
        self._task = task
        self._parsed_project = self._task.get_parsed_project()
        self._metrics = self._task.default_metrics + list(
            self._task.custom_metrics.keys()
        )
        self._parameters_structure = self._task.space.get_structure()

        problem = ApproximatedProblem(self, task)
        genetic_algorithm = self._make_genetic_algorithm(task)
        termination = self._make_genetic_setting(task.termination, get_termination)

        Thread(
            target=self._run_optimization,
            args=(problem, genetic_algorithm, termination),
        ).start()

        point = self._get_next_design_point()
        if isinstance(point, EndResult):
            raise ValueError(
                "Exploration stopped before starting, found 'End' as first point to evaluate."
            )
        self._set_point(point, self._task)

        design_point = DesignPoint.make_design_point_from_vector(
            vector=point,
            parameters_structure=self._parameters_structure,
            parsed_project=self._parsed_project,
        )

        return design_point

    def _evaluate(self, point: list[int]) -> list[float]:
        evaluation = self._get_evaluation(point)
        if evaluation is not None:
            return evaluation

        lock = Lock()
        lock.acquire()
        self._points_under_evaluation.append((point, lock))
        point_to_evaluate = [point, True]
        self._points_queue.put(point_to_evaluate)
        lock.acquire()

        point_to_evaluate[1] = False
        self._points_under_evaluation.remove((point, lock))

        evaluation = self._get_evaluation(point)

        if evaluation is None:
            raise ValueError("Cannot find evaluation for point " + str(point))

        return evaluation

    def _release(self, evaluated_design_point: EvaluatedDesignPoint):
        for (point, lock) in self._points_under_evaluation:
            if point == evaluated_design_point.vectorized():
                lock.release()
                return
        raise ValueError(
            "Could not find "
            + str(evaluated_design_point.vectorized())
            + " among points under evaluation."
        )

    def _save_results(self, result: Result):
        best_parameters_set: Optional[Iterable] = result.X
        best_metrics_set: Optional[Iterable] = result.F

        if self._task is None:
            print(best_parameters_set)
            print(best_metrics_set)
            raise ValueError("Cannot save results, found None task")

        Path(self._task.work_directory).mkdir(exist_ok=True)
        parameter_file = Path(
            self._task.work_directory,
            self._get_file_name(self._task.work_directory, self.__parameter_file_name),
        )
        metrics_file = Path(
            self._task.work_directory,
            self._get_file_name(self._task.work_directory, self.__metrics_file_name),
        )
        parameter_file.touch()
        metrics_file.touch()

        Explorer._add_row_to_output(self._task.space.get_parameters(), parameter_file)
        Explorer._add_row_to_output(self._metrics, metrics_file)

        if best_parameters_set is None or best_metrics_set is None:
            print("Optimization did not converge")
            return

        for best_parameters in best_parameters_set:
            Explorer._add_row_to_output(best_parameters, parameter_file)

        for best_metrics in best_metrics_set:
            Explorer._add_row_to_output(best_metrics, metrics_file)

    def _set_point(self, point: list[int], task: AutomaticExplorationProject) -> None:
        to_write = {}
        current_parameter = -1
        for source, module_to_parameter in self._parameters_structure.items():
            for module, parameters in module_to_parameter.items():
                to_write[module] = {}
                for parameter in parameters:
                    current_parameter += 1
                    if len(point) == current_parameter:
                        raise ValueError(
                            "Found a point dimensionally smaller than the number of parameters "
                            + str(point)
                        )
                    to_write[module][parameter] = str(point[current_parameter])

            parsed = task.parameter_sources[source]
            Path(task.project_root, source).open("w").write(parsed.replace(to_write))

    def _run_optimization(
        self, problem: Problem, algorithm: Algorithm, termination: Termination
    ) -> None:
        result = minimize(
            problem,
            algorithm,
            termination,
            seed=self.__seed,
            save_history=False,  # this cannot be turne on due to threading
            verbose=True,
        )
        self._points_queue.put(EndResult(result=result))

    def _get_next_design_point(self) -> Union[list[int], EndResult]:
        queue_item = self._points_queue.get()
        if isinstance(queue_item, EndResult):
            return queue_item
        point, to_be_evaluated = tuple(queue_item)
        while not to_be_evaluated:
            queue_item = self._points_queue.get()
            if isinstance(queue_item, EndResult):
                return queue_item
            point, to_be_evaluated = tuple(queue_item)

        if type(point) is list:
            return point
        raise ValueError(
            "Found a point " + str(point) + " which is not a list of integers."
        )

    def _make_genetic_algorithm(self, task: AutomaticExplorationProject) -> Algorithm:
        pop_size = task.options.get("pop_size")
        optional_n_offspring = task.options.get("n_offspring")
        if pop_size is None:
            raise ValueError(
                "Population size None, it is mandatory to initalize genetic algorithm"
            )

        return self._get_algorithm_class(task.algorithm)(
            repair=StepRepair(task.space.steps()),
            pop_size=pop_size * task.space.number_of_parameters(),
            n_offsprings=optional_n_offspring * task.space.number_of_parameters()
            if optional_n_offspring is not None
            else None,
            sampling=self._make_genetic_setting(task.sampling, get_sampling),
            crossover=self._make_genetic_setting(task.crossover, get_crossover),
            mutation=self._make_genetic_setting(task.mutation, get_mutation),
        )

    @staticmethod
    def _make_genetic_setting(
        settings_dict: dict[str, Any], factory: Callable[..., Any]
    ) -> Any:
        method = settings_dict.get("method")
        if method is None:
            return None
        kwargs = {k: v for k, v in settings_dict.items() if k != "method"}
        return factory(method, **kwargs)

    def _get_algorithm_class(self, algorithm_name: str) -> type[Algorithm]:
        algorithm_module = importlib.import_module(
            self.__pymoo_package_structure + "." + algorithm_name.lower()
            if algorithm_name.lower() != "agemoea"
            else "age"
        )
        return getattr(algorithm_module, algorithm_name.upper())

    def _get_evaluation(self, point: list[int]) -> Optional[list[float]]:
        for stored_point, evaluation in self._evaluated_points:
            if stored_point == point:
                self._evaluated_points.remove((stored_point, evaluation))
                return [evaluation.design_value[metric] for metric in self._metrics]
        return None


class ApproximatedProblem(ElementwiseProblem):
    def __init__(self, explorer: AutomaticExplorer, task: AutomaticExplorationProject):
        ranges = [range for _, range in task.space.get_ranges()]
        self._explorer = explorer
        self._objectives = len(task.default_metrics) + len(task.custom_metrics)
        self._fitness = approximate(
            outputs=self._objectives,
            controller=task.approximation_options["controller"],
            estimator=task.approximation_options["estimator"],
            disabled=not task.approximate,
            debug=True,
        )(self._explorer._evaluate)
        super().__init__(
            n_var=task.space.number_of_parameters(),
            n_obj=self._objectives,
            xl=[range.start for range in ranges],
            xu=[range.end for range in ranges],
            type_var=np.int_,
        )

    def _evaluate(self, x: Iterable[int], out, **kwargs) -> None:
        # kwargs accomodates pymoo abstraction, do not remove
        fitness = self._fitness(list(x))
        out["F"] = fitness


class StepRepair(Repair):
    def __init__(self, repairs: list[Union[int, Literal["powers_of_two"]]]) -> None:
        self.__repairs = repairs
        super().__init__()

    def _do(self, problem, pop, **kwargs):
        # The signature accomodates pymoo abstraction do not remove unused parameters
        for k, _ in enumerate(pop):
            point = pop[k].X
            for parameter, repair in enumerate(self.__repairs):
                if repair == "powers_of_two":
                    point[parameter] = self._repair_to_power_of_two(point[parameter])
                else:
                    point[parameter] = self._repair_to_step(point[parameter], repair)
        return pop

    @staticmethod
    def _repair_to_step(value_to_repair: int, step: int):
        if step <= 1:
            return value_to_repair

        larger_multiple: int = np.floor(value_to_repair / step) * step
        smaller_multiple: int = np.ceil(value_to_repair / step) * step
        larger_to_value_distance: int = larger_multiple - value_to_repair
        smaller_to_value_distance: int = value_to_repair - smaller_multiple

        if larger_to_value_distance == smaller_to_value_distance:
            if np.random.random() < 0.5:
                return smaller_multiple

            return larger_multiple

        if larger_to_value_distance < smaller_to_value_distance:
            return larger_multiple

        return smaller_multiple

    @staticmethod
    def _repair_to_power_of_two(value_to_repair: int):
        next_power_of_two = value_to_repair
        next_power_of_two -= 1
        next_power_of_two |= next_power_of_two >> 1
        next_power_of_two |= next_power_of_two >> 2
        next_power_of_two |= next_power_of_two >> 4
        next_power_of_two |= next_power_of_two >> 8
        next_power_of_two |= next_power_of_two >> 16
        next_power_of_two += 1

        previous_power_of_two = next_power_of_two >> 1
        if np.random.random() < 0.5:
            return previous_power_of_two
        else:
            return next_power_of_two



from dovado_rtl.parsers.utilities.antlr.hdl.hdl_antlr_parsed import HdlAntlrParsed


class VerilogParsed(HdlAntlrParsed):
    @property
    def _parameter_initialization_prefix(self) -> str:
        raise Exception("Parameters in Verilog must be initialized")



from dovado_rtl.generators.generator import Generator
from dovado_rtl.explorers.utilities.design_points import DesignPoint
import subprocess
from pathlib import Path
from importlib.resources import files
from typing import Optional
from nacolla.models import ImmutableModel


class VitisProject(ImmutableModel):
    project_root: str
    target_file: str
    target_module: str
    clock_port: Optional[str] = None
    target_clock: Optional[float] = None


class VitisGenerator(Generator):
    _tcl_file_name = "csynth.tcl"
    _generated_folder_name = "hls_generated"

    def generate(self, design_point: DesignPoint) -> DesignPoint:
        if design_point.extra is None or design_point.extra.get("vitis") is None:
            raise ValueError(
                "Please provide an extra.vitis section in the configuration file"
            )
        vitis_project = VitisProject(**design_point.extra["vitis"])

        target_path = Path(
            design_point.project_root, design_point.target_file
        ).relative_to(design_point.work_directory)

        tcl_path = str(
            files("dovado_rtl.generators.vitis").joinpath(VitisGenerator._tcl_file_name)
        )

        target_clock: float = (
            vitis_project.target_clock
            if vitis_project.target_clock is not None
            else design_point.target_clock
        )

        vitis_subcommands: list[str] = [
            "vitis_hls " + str(tcl_path),
            "-tclargs",
            "empty",
            "empty",
            VitisGenerator._generated_folder_name,  # output folder
            str(target_path),
            design_point.board,
            format(1000 / target_clock, "f"),  # target clock in nanoseconds
            vitis_project.target_module,
            str(
                design_point.project_root.relative_to(design_point.work_directory)
            ),  # directory to include
        ]

        vitis_command = " ".join(vitis_subcommands)

        vitis = subprocess.Popen(
            vitis_command,
            shell=True,
            stdout=subprocess.PIPE,
            cwd=design_point.work_directory,
        )
        vitis.wait()

        vitis_project_no_clock = {
            k: v for k, v in dict(vitis_project).items() if k != "target_clock"
        }

        generated_project_root = Path(
            design_point.work_directory,
            VitisGenerator._generated_folder_name,
            vitis_project.project_root,
        )
        return DesignPoint(
            **(
                dict(design_point)
                | vitis_project_no_clock
                | {"project_root": generated_project_root}
            )
        )



from abc import ABC, abstractmethod
from typing import Sequence
from antlr4 import TokenStream
from antlr4.TokenStreamRewriter import TokenStreamRewriter
from dovado_rtl.parsers.utilities.antlr.antlr_module import AntlrModule
from dovado_rtl.parsers.utilities.antlr.antlr_parameter import AntlrParameter
from dovado_rtl.parsers.utilities.parsed import (
    MODULE_NAME_STR,
    PARAMETER_NAME_STR,
    VALUE_STR,
    Parsed,
)


class AntlrParsed(ABC, Parsed):
    def __init__(
        self, token_stream: TokenStream, modules: Sequence[AntlrModule]
    ) -> None:
        self._modules = modules
        self.__rewriter = TokenStreamRewriter(token_stream)

    @property
    def modules(self) -> Sequence[AntlrModule]:
        return self._modules

    @property
    @abstractmethod
    def _parameter_initialization_prefix(self) -> str:
        ...

    def get_module(self, name: str) -> AntlrModule:
        if "/" in name:
            raise ValueError(
                "Module identifiers containing '/' are not supported.\n"
                + "In the future a module name containing '/' will represent a path to a submodule.\n"
                + " Submodules are not supported for now."
            )

        for module in self.modules:
            if name == module.name:
                return module
        raise ValueError(
            "Module '"
            + str(name)
            + "' does not exist.\n"
            + "WARNING: module identifiers containing '/' are not supported."
        )

    def _get_parameter(self, parameter_name: str, module_name: str) -> AntlrParameter:
        module = self.get_module(module_name)
        for parameter in module.parameters:
            if parameter.name == parameter_name:
                return parameter
        raise ValueError(
            "Parameter '"
            + str(parameter_name)
            + "' does not exist in module '"
            + str(module_name)
            + "'"
        )

    def _lazy_replace(self, parameter: AntlrParameter, value: str) -> None:
        parameter.value = value

        to_write = value
        if not parameter.has_default:
            to_write = (
                parameter.rule.getText() + self._parameter_initialization_prefix + value
            )

        self.__rewriter.replaceRangeTokens(
            parameter.rule.start, parameter.rule.stop, to_write
        )

    def replace(
        self,
        module_parameter_to_value: dict[
            MODULE_NAME_STR, dict[PARAMETER_NAME_STR, VALUE_STR]
        ],
    ) -> str:
        for module_name, parameter_to_value in module_parameter_to_value.items():
            for parameter_name, value in parameter_to_value.items():
                self._lazy_replace(
                    self._get_parameter(parameter_name, module_name), value
                )
        return self.__rewriter.getDefaultText()



# Generated from VerilogPreParser.g4 by ANTLR 4.11.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .VerilogPreParser import VerilogPreParser
else:
    from VerilogPreParser import VerilogPreParser

# This class defines a complete listener for a parse tree produced by VerilogPreParser.
class VerilogPreParserListener(ParseTreeListener):

    # Enter a parse tree produced by VerilogPreParser#source_text.
    def enterSource_text(self, ctx:VerilogPreParser.Source_textContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#source_text.
    def exitSource_text(self, ctx:VerilogPreParser.Source_textContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#compiler_directive.
    def enterCompiler_directive(self, ctx:VerilogPreParser.Compiler_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#compiler_directive.
    def exitCompiler_directive(self, ctx:VerilogPreParser.Compiler_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#begin_keywords_directive.
    def enterBegin_keywords_directive(self, ctx:VerilogPreParser.Begin_keywords_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#begin_keywords_directive.
    def exitBegin_keywords_directive(self, ctx:VerilogPreParser.Begin_keywords_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#celldefine_directive.
    def enterCelldefine_directive(self, ctx:VerilogPreParser.Celldefine_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#celldefine_directive.
    def exitCelldefine_directive(self, ctx:VerilogPreParser.Celldefine_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#default_nettype_directive.
    def enterDefault_nettype_directive(self, ctx:VerilogPreParser.Default_nettype_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#default_nettype_directive.
    def exitDefault_nettype_directive(self, ctx:VerilogPreParser.Default_nettype_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#default_nettype_value.
    def enterDefault_nettype_value(self, ctx:VerilogPreParser.Default_nettype_valueContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#default_nettype_value.
    def exitDefault_nettype_value(self, ctx:VerilogPreParser.Default_nettype_valueContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#else_directive.
    def enterElse_directive(self, ctx:VerilogPreParser.Else_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#else_directive.
    def exitElse_directive(self, ctx:VerilogPreParser.Else_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#elsif_directive.
    def enterElsif_directive(self, ctx:VerilogPreParser.Elsif_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#elsif_directive.
    def exitElsif_directive(self, ctx:VerilogPreParser.Elsif_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#end_keywords_directive.
    def enterEnd_keywords_directive(self, ctx:VerilogPreParser.End_keywords_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#end_keywords_directive.
    def exitEnd_keywords_directive(self, ctx:VerilogPreParser.End_keywords_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#endcelldefine_directive.
    def enterEndcelldefine_directive(self, ctx:VerilogPreParser.Endcelldefine_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#endcelldefine_directive.
    def exitEndcelldefine_directive(self, ctx:VerilogPreParser.Endcelldefine_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#endif_directive.
    def enterEndif_directive(self, ctx:VerilogPreParser.Endif_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#endif_directive.
    def exitEndif_directive(self, ctx:VerilogPreParser.Endif_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#filename.
    def enterFilename(self, ctx:VerilogPreParser.FilenameContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#filename.
    def exitFilename(self, ctx:VerilogPreParser.FilenameContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#group_of_lines.
    def enterGroup_of_lines(self, ctx:VerilogPreParser.Group_of_linesContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#group_of_lines.
    def exitGroup_of_lines(self, ctx:VerilogPreParser.Group_of_linesContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#identifier.
    def enterIdentifier(self, ctx:VerilogPreParser.IdentifierContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#identifier.
    def exitIdentifier(self, ctx:VerilogPreParser.IdentifierContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#ifdef_directive.
    def enterIfdef_directive(self, ctx:VerilogPreParser.Ifdef_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#ifdef_directive.
    def exitIfdef_directive(self, ctx:VerilogPreParser.Ifdef_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#ifndef_directive.
    def enterIfndef_directive(self, ctx:VerilogPreParser.Ifndef_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#ifndef_directive.
    def exitIfndef_directive(self, ctx:VerilogPreParser.Ifndef_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#include_directive.
    def enterInclude_directive(self, ctx:VerilogPreParser.Include_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#include_directive.
    def exitInclude_directive(self, ctx:VerilogPreParser.Include_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#level.
    def enterLevel(self, ctx:VerilogPreParser.LevelContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#level.
    def exitLevel(self, ctx:VerilogPreParser.LevelContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#line_directive.
    def enterLine_directive(self, ctx:VerilogPreParser.Line_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#line_directive.
    def exitLine_directive(self, ctx:VerilogPreParser.Line_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#macro_delimiter.
    def enterMacro_delimiter(self, ctx:VerilogPreParser.Macro_delimiterContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#macro_delimiter.
    def exitMacro_delimiter(self, ctx:VerilogPreParser.Macro_delimiterContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#macro_esc_newline.
    def enterMacro_esc_newline(self, ctx:VerilogPreParser.Macro_esc_newlineContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#macro_esc_newline.
    def exitMacro_esc_newline(self, ctx:VerilogPreParser.Macro_esc_newlineContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#macro_esc_quote.
    def enterMacro_esc_quote(self, ctx:VerilogPreParser.Macro_esc_quoteContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#macro_esc_quote.
    def exitMacro_esc_quote(self, ctx:VerilogPreParser.Macro_esc_quoteContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#macro_identifier.
    def enterMacro_identifier(self, ctx:VerilogPreParser.Macro_identifierContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#macro_identifier.
    def exitMacro_identifier(self, ctx:VerilogPreParser.Macro_identifierContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#macro_name.
    def enterMacro_name(self, ctx:VerilogPreParser.Macro_nameContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#macro_name.
    def exitMacro_name(self, ctx:VerilogPreParser.Macro_nameContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#macro_quote.
    def enterMacro_quote(self, ctx:VerilogPreParser.Macro_quoteContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#macro_quote.
    def exitMacro_quote(self, ctx:VerilogPreParser.Macro_quoteContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#macro_text.
    def enterMacro_text(self, ctx:VerilogPreParser.Macro_textContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#macro_text.
    def exitMacro_text(self, ctx:VerilogPreParser.Macro_textContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#macro_usage.
    def enterMacro_usage(self, ctx:VerilogPreParser.Macro_usageContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#macro_usage.
    def exitMacro_usage(self, ctx:VerilogPreParser.Macro_usageContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#nounconnected_drive_directive.
    def enterNounconnected_drive_directive(self, ctx:VerilogPreParser.Nounconnected_drive_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#nounconnected_drive_directive.
    def exitNounconnected_drive_directive(self, ctx:VerilogPreParser.Nounconnected_drive_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#number.
    def enterNumber(self, ctx:VerilogPreParser.NumberContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#number.
    def exitNumber(self, ctx:VerilogPreParser.NumberContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#pragma_directive.
    def enterPragma_directive(self, ctx:VerilogPreParser.Pragma_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#pragma_directive.
    def exitPragma_directive(self, ctx:VerilogPreParser.Pragma_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#pragma_expression.
    def enterPragma_expression(self, ctx:VerilogPreParser.Pragma_expressionContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#pragma_expression.
    def exitPragma_expression(self, ctx:VerilogPreParser.Pragma_expressionContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#pragma_keyword.
    def enterPragma_keyword(self, ctx:VerilogPreParser.Pragma_keywordContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#pragma_keyword.
    def exitPragma_keyword(self, ctx:VerilogPreParser.Pragma_keywordContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#pragma_name.
    def enterPragma_name(self, ctx:VerilogPreParser.Pragma_nameContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#pragma_name.
    def exitPragma_name(self, ctx:VerilogPreParser.Pragma_nameContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#pragma_value.
    def enterPragma_value(self, ctx:VerilogPreParser.Pragma_valueContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#pragma_value.
    def exitPragma_value(self, ctx:VerilogPreParser.Pragma_valueContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#resetall_directive.
    def enterResetall_directive(self, ctx:VerilogPreParser.Resetall_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#resetall_directive.
    def exitResetall_directive(self, ctx:VerilogPreParser.Resetall_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#source_text_.
    def enterSource_text_(self, ctx:VerilogPreParser.Source_text_Context):
        pass

    # Exit a parse tree produced by VerilogPreParser#source_text_.
    def exitSource_text_(self, ctx:VerilogPreParser.Source_text_Context):
        pass


    # Enter a parse tree produced by VerilogPreParser#string_.
    def enterString_(self, ctx:VerilogPreParser.String_Context):
        pass

    # Exit a parse tree produced by VerilogPreParser#string_.
    def exitString_(self, ctx:VerilogPreParser.String_Context):
        pass


    # Enter a parse tree produced by VerilogPreParser#text_macro_definition.
    def enterText_macro_definition(self, ctx:VerilogPreParser.Text_macro_definitionContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#text_macro_definition.
    def exitText_macro_definition(self, ctx:VerilogPreParser.Text_macro_definitionContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#text_macro_usage.
    def enterText_macro_usage(self, ctx:VerilogPreParser.Text_macro_usageContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#text_macro_usage.
    def exitText_macro_usage(self, ctx:VerilogPreParser.Text_macro_usageContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#time_precision.
    def enterTime_precision(self, ctx:VerilogPreParser.Time_precisionContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#time_precision.
    def exitTime_precision(self, ctx:VerilogPreParser.Time_precisionContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#time_unit.
    def enterTime_unit(self, ctx:VerilogPreParser.Time_unitContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#time_unit.
    def exitTime_unit(self, ctx:VerilogPreParser.Time_unitContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#timescale_directive.
    def enterTimescale_directive(self, ctx:VerilogPreParser.Timescale_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#timescale_directive.
    def exitTimescale_directive(self, ctx:VerilogPreParser.Timescale_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#unconnected_drive_directive.
    def enterUnconnected_drive_directive(self, ctx:VerilogPreParser.Unconnected_drive_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#unconnected_drive_directive.
    def exitUnconnected_drive_directive(self, ctx:VerilogPreParser.Unconnected_drive_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#unconnected_drive_value.
    def enterUnconnected_drive_value(self, ctx:VerilogPreParser.Unconnected_drive_valueContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#unconnected_drive_value.
    def exitUnconnected_drive_value(self, ctx:VerilogPreParser.Unconnected_drive_valueContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#undef_directive.
    def enterUndef_directive(self, ctx:VerilogPreParser.Undef_directiveContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#undef_directive.
    def exitUndef_directive(self, ctx:VerilogPreParser.Undef_directiveContext):
        pass


    # Enter a parse tree produced by VerilogPreParser#version_specifier.
    def enterVersion_specifier(self, ctx:VerilogPreParser.Version_specifierContext):
        pass

    # Exit a parse tree produced by VerilogPreParser#version_specifier.
    def exitVersion_specifier(self, ctx:VerilogPreParser.Version_specifierContext):
        pass



del VerilogPreParser


from collections import deque
from pathlib import Path
from typing import Optional, Union

from dovado_rtl.explorers.explorer import (
    DesignPoint,
    EndExploration,
    EvaluatedDesignPoint,
    Explorer,
)
from dovado_rtl.explorers.utilities.spaces import SOURCE_PATH
from dovado_rtl.explorers.utilities.tasks import ManualExplorationProject
from dovado_rtl.parsers.utilities.parsed import MODULE_NAME_STR, PARAMETER_NAME_STR


class ManualExplorer(Explorer):
    _output_file_name = "exploration"
    _homonymous_files_limit = 100

    def __init__(self) -> None:
        super().__init__()
        self._evaluated_points = 0
        self._output_file: Optional[Path] = None
        self._task: Optional[ManualExplorationProject] = None
        self._metrics: list[str] = []
        self._points_under_evaluation: deque[DesignPoint] = deque()
        self._buffered_evaluated_points: list[EvaluatedDesignPoint] = []

    def explore(self, task: ManualExplorationProject) -> DesignPoint:
        self._task = task
        self._metrics = task.default_metrics + list(task.custom_metrics.keys())

        Path(task.work_directory).mkdir(exist_ok=True)
        self._output_file = Path(
            task.work_directory,
            self._get_file_name(
                self._task.work_directory, ManualExplorer._output_file_name
            ),
        )
        self._output_file.touch(exist_ok=False)

        Explorer._add_row_to_output(self._metrics, self._output_file)

        current_points = self._set_point(self._evaluated_points, task)
        if current_points is None:
            raise ValueError("Found at least one parameter with no values to explore")

        design_point = DesignPoint(
            **(dict(self._task.get_parsed_project()) | {"points": current_points})
        )
        self._points_under_evaluation.append(design_point)
        return design_point

    def update(
        self, evaluated_design_point: Union[EvaluatedDesignPoint, EndExploration]
    ) -> Union[DesignPoint, EndExploration]:
        if isinstance(evaluated_design_point, EndExploration):
            return EndExploration()

        if self._task is None:
            raise ValueError("Found None task in manual explorer")
        if self._output_file is None:
            raise ValueError("Found None output_file")

        if self._points_under_evaluation[0].points == evaluated_design_point.points:
            Explorer._add_row_to_output(
                [
                    str(evaluated_design_point.design_value[metric])
                    for metric in self._metrics
                ],
                self._output_file,
            )
            self._points_under_evaluation.pop()
            self._add_buffered_points_to_output()

        else:
            self._buffered_evaluated_points.append(evaluated_design_point)

        current_points = self._set_point(self._evaluated_points, self._task)
        if current_points is None:
            return EndExploration()

        design_point = DesignPoint(
            **(dict(self._task.get_parsed_project()) | {"points": current_points})
        )
        self._points_under_evaluation.append(design_point)
        return design_point

    def _add_buffered_points_to_output(self) -> None:
        if self._output_file is None:
            raise ValueError("Found None output file")
        for point_under_evaluation in self._points_under_evaluation:
            used_buffered_point = False
            for buffered_evaluated_point in self._buffered_evaluated_points:
                if point_under_evaluation.points == buffered_evaluated_point.points:
                    Explorer._add_row_to_output(
                        [
                            str(buffered_evaluated_point.design_value[metric])
                            for metric in self._metrics
                        ],
                        self._output_file,
                    )
                    self._points_under_evaluation.pop()
                    self._buffered_evaluated_points.remove(buffered_evaluated_point)
                    used_buffered_point = True
                    break
            if not used_buffered_point:
                break

    def _set_point(
        self, iterations: int, task: ManualExplorationProject
    ) -> Optional[
        dict[SOURCE_PATH, dict[MODULE_NAME_STR, dict[PARAMETER_NAME_STR, str]]]
    ]:
        current_points = {}
        for source_path, modules_to_parameters_dict in task.space.points.items():
            parsed = task.parameter_sources[source_path]
            current_points[source_path] = {}
            module_to_parameter_to_value = {}

            for module_name, parameters_to_values in modules_to_parameters_dict.items():
                module_to_parameter_to_value[module_name] = {}

                for parameter, values in parameters_to_values.items():
                    if len(values) == iterations:
                        return None

                    module_to_parameter_to_value[module_name][parameter] = values[
                        iterations
                    ]

                current_points[source_path].update(module_to_parameter_to_value)

            Path(task.project_root, source_path).open("w").write(
                parsed.replace(module_to_parameter_to_value)
            )

        self._evaluated_points += 1
        return current_points



from typing import Protocol

MODULE_NAME_STR = str
PARAMETER_NAME_STR = str
VALUE_STR = str

from typing_extensions import runtime_checkable


@runtime_checkable
class Parsed(Protocol):
    def replace(
        self,
        parameter_to_value: dict[MODULE_NAME_STR, dict[PARAMETER_NAME_STR, VALUE_STR]],
    ) -> str:
        ...



from typing import Optional, Sequence, Union

from antlr4 import IllegalStateException
from dovado_rtl.parsers.vhdl.vhdl_parameter import VhdlParameter
from dovado_rtl.parsers.utilities import Port, HdlAntlrModule
from dovado_rtl.parsers.vhdl.generated.vhdlVisitor import vhdlVisitor
from dovado_rtl.parsers.vhdl.generated.vhdlParser import vhdlParser
from dovado_rtl.parsers.utilities.port import PORT_DIMENSION, PORT_DIRECTION


class VhdlVisitor(vhdlVisitor):
    def visitDesign_file(
        self, ctx: vhdlParser.Design_fileContext
    ) -> Sequence[HdlAntlrModule]:
        modules: list[HdlAntlrModule]

        design_units = ctx.design_unit()

        modules = []
        if design_units:
            for design_unit in design_units:
                module = self.visitDesign_unit(design_unit)
                if module is not None:
                    modules.append(module)

        return modules

    def visitDesign_unit(
        self, ctx: vhdlParser.Design_unitContext
    ) -> Optional[HdlAntlrModule]:
        # context clause is ignored
        library_unit = ctx.library_unit()

        if library_unit:
            return self.visitLibrary_unit(library_unit)
        return None

    def visitLibrary_unit(
        self, ctx: vhdlParser.Library_unitContext
    ) -> Optional[HdlAntlrModule]:
        primary_unit = ctx.primary_unit()

        if primary_unit:
            return self.visitPrimary_unit(primary_unit)
        return None

    def visitPrimary_unit(
        self, ctx: vhdlParser.Primary_unitContext
    ) -> Optional[HdlAntlrModule]:
        # packages and configurations are ignored
        entity_declaration = ctx.entity_declaration()

        if entity_declaration:
            return self.visitEntity_declaration(entity_declaration)
        return None

    def visitEntity_declaration(self, ctx: vhdlParser.Entity_declarationContext):
        parsed_identifier: str
        parameters: list[VhdlParameter]
        ports: list[Port]

        identifiers = ctx.identifier()
        entity_header = ctx.entity_header()

        if identifiers:
            # skipping the second name declaration
            parsed_identifier = identifiers[0].getText()
            if entity_header:
                parameters, ports = self.visitEntity_header(entity_header)
                return HdlAntlrModule(
                    name=parsed_identifier, parameters=parameters, ports=ports
                )
            else:
                raise IllegalStateException("Entity must have a header section")

        else:
            raise IllegalStateException("Identifier is mandatory for an entity")

    def visitEntity_header(
        self, ctx: vhdlParser.Entity_headerContext
    ) -> tuple[list[VhdlParameter], list[Port]]:
        ports: list[Port]
        parameters: list[VhdlParameter]

        generic_clause = ctx.generic_clause()
        port_clause = ctx.port_clause()

        ports = []
        parameters = []
        if generic_clause:
            parameters = self.visitGeneric_clause(generic_clause)
        if port_clause:
            ports = self.visitPort_clause(port_clause)
        return parameters, ports

    def visitGeneric_clause(
        self, ctx: vhdlParser.Generic_clauseContext
    ) -> list[VhdlParameter]:
        generic_list = ctx.generic_list()

        if generic_list:
            return self.visitGeneric_list(generic_list)
        return []

    def visitPort_clause(self, ctx: vhdlParser.Port_clauseContext) -> list[Port]:
        port_list = ctx.port_list()
        if port_list:
            return self.visitPort_list(port_list)
        return []

    def visitPort_list(self, ctx: vhdlParser.Port_listContext) -> list[Port]:
        interface_port_list = ctx.interface_port_list()

        if interface_port_list:
            return self.visitInterface_port_list(interface_port_list)
        return []

    def visitInterface_port_list(
        self, ctx: vhdlParser.Interface_port_listContext
    ) -> list[Port]:
        ports: list[Port]

        interface_port_declarations = ctx.interface_port_declaration()

        ports = []
        if interface_port_declarations:
            for interface_port_declaration in interface_port_declarations:
                ports += self.visitInterface_port_declaration(
                    interface_port_declaration
                )
        return ports

    def visitInterface_port_declaration(
        self, ctx: vhdlParser.Interface_port_declarationContext
    ) -> list[Port]:
        identifiers: list[str]
        port_direction: PORT_DIRECTION
        port_dimension: PORT_DIMENSION

        identifier_list = ctx.identifier_list()
        signal_mode = ctx.signal_mode()
        subtype_indication = ctx.subtype_indication()
        expression = ctx.expression()

        if identifier_list:
            identifiers = self.visitIdentifier_list(identifier_list)
            if signal_mode:
                port_direction = self.visitSignal_mode(signal_mode)
            else:
                # if no direction is specified
                # the standard assumes input
                port_direction = "input"

            if subtype_indication:
                _, port_dimension = self.visitSubtype_indication(subtype_indication)
            else:
                raise IllegalStateException("Type is mandatory in port declaration")

            return [
                Port(
                    name=identifier,
                    direction=port_direction,
                    dimension=port_dimension,
                    has_default=expression is not None,
                )
                for identifier in identifiers
            ]
        return []

    @staticmethod
    def is_subtype_scalar_boolean(selected_names: list[str]) -> bool:
        if len(selected_names) > 1:
            return False
        return selected_names[0] == "boolean"

    def visitSubtype_indication(
        self, ctx: vhdlParser.Subtype_indicationContext
    ) -> tuple[bool, PORT_DIMENSION]:
        parsed_selected_names: list[str]

        selected_names = ctx.selected_name()
        if selected_names:
            parsed_selected_names = [
                self.visitSelected_name(selected_name)
                for selected_name in selected_names
            ]
        else:
            raise IllegalStateException(
                "A subtype indication must contain at least one type"
            )

        if ctx.constraint():
            return False, "vectorial"
        else:
            return self.is_subtype_scalar_boolean(parsed_selected_names), "scalar"

    def visitSelected_name(self, ctx: vhdlParser.Selected_nameContext) -> str:
        identifier = ctx.identifier()
        if identifier:
            return self.visitIdentifier(identifier)
        raise IllegalStateException("A selected name must contain an identifier")

    def visitIdentifier(self, ctx: vhdlParser.IdentifierContext) -> str:
        basic_identifier = ctx.BASIC_IDENTIFIER()
        extended_identifier = ctx.EXTENDED_IDENTIFIER()
        if basic_identifier:
            return str(basic_identifier)
        if extended_identifier:
            return str(extended_identifier)
        raise IllegalStateException(
            "An identifier must contain either a base identifier or an extended identifier"
        )

    def visitSignal_mode(self, ctx: vhdlParser.Signal_modeContext) -> PORT_DIRECTION:
        if ctx.IN():
            return "input"
        if ctx.OUT():
            return "output"
        if ctx.INOUT():
            return "inout"
        if ctx.BUFFER():
            return "input"
        if ctx.LINKAGE():
            return "input"
        raise IllegalStateException("Found illegal signal mode")

    def visitGeneric_list(self, ctx: vhdlParser.Generic_listContext):
        parameters: list[VhdlParameter]
        interface_constant_declarations = ctx.interface_constant_declaration()

        parameters = []
        if interface_constant_declarations:
            for interface_constant_declaration in interface_constant_declarations:
                parameters += self.visitInterface_constant_declaration(
                    interface_constant_declaration
                )

        return parameters

    def visitInterface_constant_declaration(
        self, ctx: vhdlParser.Interface_constant_declarationContext
    ) -> list[VhdlParameter]:
        identifiers: list[str]
        value: str
        rule: Union[vhdlParser.ExpressionContext, vhdlParser.Identifier_listContext]

        identifier_list = ctx.identifier_list()
        expression = ctx.expression()
        subtype_indication = ctx.subtype_indication()

        if not subtype_indication:
            raise IllegalStateException("Generics must be strictly typed")

        if identifier_list:
            identifiers = self.visitIdentifier_list(identifier_list)

            value = ""
            if expression:
                value = expression.getText()
                rule = expression
            else:
                rule = subtype_indication

            return [
                VhdlParameter(
                    name=identifier,
                    value=value,
                    rule=rule,
                    has_default=True if value != "" else False,
                    is_boolean=self.visitSubtype_indication(subtype_indication)[0],
                )
                for identifier in identifiers
            ]
        return []

    def visitIdentifier_list(self, ctx: vhdlParser.Identifier_listContext) -> list[str]:
        identifiers = ctx.identifier()

        if identifiers:
            return [identifier.getText() for identifier in identifiers]

        return []



from dovado_rtl.explorers.utilities.design_points import DesignPoint
from nacolla.stateful_callable import StatefulCallable
from abc import abstractmethod


class Generator(StatefulCallable):
    @abstractmethod
    def generate(self, design_point: DesignPoint) -> DesignPoint:
        ...



from typing import Any, Optional, Sequence, Union, cast
from antlr4 import IllegalStateException
from dovado_rtl.parsers.system_verilog.generated.SystemVerilogParserVisitor import (
    SystemVerilogParserVisitor,
)
from dovado_rtl.parsers.system_verilog.generated.SystemVerilogParser import (
    SystemVerilogParser,
)
from dovado_rtl.parsers.utilities import (
    AntlrParameter,
    HdlAntlrModule,
    PORT_DIRECTION,
    PORT_DIMENSION,
    Port,
)


class SystemVerilogVisitor(SystemVerilogParserVisitor):
    def visitSource_text(
        self, ctx: SystemVerilogParser.Source_textContext
    ) -> Sequence[HdlAntlrModule]:
        modules: list[HdlAntlrModule]

        descriptions = ctx.description()

        modules = []
        if descriptions:
            for description in descriptions:
                modules.append(self.visitDescription(description))
        return modules

    def visitDescription(
        self, ctx: SystemVerilogParser.DescriptionContext
    ) -> HdlAntlrModule:
        # configurations, bind directives, attributes and udp are ignored
        module_declaration = ctx.module_declaration()
        interface_declaration = ctx.interface_declaration()
        package_declaration = ctx.package_declaration()
        package_item = ctx.package_item()

        if module_declaration:
            return self.visitModule_declaration(module_declaration)
        elif interface_declaration:
            return self.visitInterface_declaration(interface_declaration)
        elif package_declaration:
            return self.visitPackage_declaration(package_declaration)
        elif package_item:
            return HdlAntlrModule(
                name="", parameters=self.visitPackage_item(package_item), ports=[]
            )
        else:
            raise IllegalStateException("Empty description found.")

    def visitPackage_declaration(
        self, ctx: SystemVerilogParser.Package_declarationContext
    ) -> HdlAntlrModule:
        parsed_identifier: str
        parameters: list[AntlrParameter]

        package_identifier = ctx.package_identifier()
        pkg_decl_items = ctx.pkg_decl_item()

        if package_identifier:
            parsed_identifier = package_identifier.getText()
        else:
            raise IllegalStateException("Package name is mandatory")
        parameters = []
        if pkg_decl_items:
            for pkg_decl_item in pkg_decl_items:
                parameters += self.visitPkg_decl_item(pkg_decl_item)
        return HdlAntlrModule(name=parsed_identifier, parameters=parameters, ports=[])

    def visitPkg_decl_item(
        self, ctx: SystemVerilogParser.Pkg_decl_itemContext
    ) -> list[AntlrParameter]:
        package_item = ctx.package_item()
        if package_item:
            return self.visitPackage_item(package_item)
        return []

    def visitPackage_item(
        self, ctx: SystemVerilogParser.Package_itemContext
    ) -> list[AntlrParameter]:
        package_item_declaration = ctx.package_item_declaration()
        if package_item_declaration:
            return self.visitPackage_item_declaration(package_item_declaration)
        return []

    def visitModule_declaration(
        self, ctx: SystemVerilogParser.Module_declarationContext
    ) -> HdlAntlrModule:
        # Module name after "endmodule" is ignored
        def parse_module_items(
            module_items: list[SystemVerilogParser.Module_itemContext],
        ) -> tuple[list[Port], list[AntlrParameter], list[HdlAntlrModule]]:
            ports = []
            parameters = []
            submodules = []
            for module_item in module_items:
                module_item = self.visitModule_item(module_item)
                if module_item:
                    if isinstance(module_item, list):
                        if isinstance(module_item[0], Port):
                            module_item = cast(list[Port], module_item)
                            ports += module_item
                        else:
                            module_item = cast(list[AntlrParameter], module_item)
                            parameters += module_item
                    else:
                        submodules.append(module_item)
                return ports, parameters, submodules
            return [], [], []

        parsed_module: HdlAntlrModule
        ports: list[Port]
        parameters: list[AntlrParameter]
        submodules: list[HdlAntlrModule]

        module_header = ctx.module_header()
        module_items = ctx.module_item()
        module_identifier = ctx.module_identifier()

        ports = []
        parameters = []
        submodules = []
        if module_header:
            parsed_module = self.visitModule_header(module_header)
            if module_items:
                ports, parameters, submodules = parse_module_items(module_items)
            return HdlAntlrModule(
                name=parsed_module.name,
                parameters=parsed_module.parameters + parameters,
                ports=parsed_module.ports + ports,
                submodules=submodules,
            )

        elif module_identifier:
            parsed_module_identifier = self.visitModule_identifier(module_identifier)
            if module_items:
                ports, parameters, submodules = parse_module_items(module_items)
            return HdlAntlrModule(
                name=parsed_module_identifier,
                parameters=parameters,
                ports=ports,
                submodules=submodules,
            )

        else:
            raise IllegalStateException(
                "Module header or module identifier are mandatory in module_declaration"
            )

    def visitModule_header(
        self, ctx: SystemVerilogParser.Module_headerContext
    ) -> HdlAntlrModule:
        parsed_module_identifier: str
        parameters: list[AntlrParameter]

        module_identifier = ctx.module_identifier()
        parameter_port_list = ctx.parameter_port_list()
        list_of_port_declarations = ctx.list_of_port_declarations()

        if module_identifier:
            parsed_module_identifier: str = self.visitModule_identifier(
                module_identifier
            )
        else:
            raise IllegalStateException(
                "Module identifier is mandatory in module header"
            )

        parameters = []
        if parameter_port_list:
            parameters = self.visitParameter_port_list(parameter_port_list)

        ports = []
        if list_of_port_declarations:
            ports = self.visitList_of_port_declarations(list_of_port_declarations)

        return HdlAntlrModule(
            name=parsed_module_identifier, parameters=parameters, ports=ports
        )

    def visitList_of_port_declarations(
        self, ctx: SystemVerilogParser.List_of_port_declarationsContext
    ) -> list[Port]:
        # we only parse port declarations
        ports: list[Port]

        port_decls = ctx.port_decl()

        if port_decls:
            ports = []
            for port_decl in port_decls:
                parsed_port_decl = self.visitPort_decl(port_decl)
                if parsed_port_decl:
                    ports.append(parsed_port_decl)
            return ports

        return []

    def visitPort_decl(
        self, ctx: SystemVerilogParser.Port_declContext
    ) -> Optional[Port]:
        # attributes are ignored
        ansi_port_declaration = ctx.ansi_port_declaration()

        if ansi_port_declaration:
            return self.visitAnsi_port_declaration(ansi_port_declaration)
        raise IllegalStateException(
            "A port decl requires a corresponding ansi port declaration"
        )

    def visitAnsi_port_declaration(
        self, ctx: SystemVerilogParser.Ansi_port_declarationContext
    ) -> Optional[Port]:
        # all ports without explicit direction are ignored
        parsed_port_direction: PORT_DIRECTION
        parsed_port_identifier: str
        port_dimension: PORT_DIMENSION

        port_direction = ctx.port_direction()
        port_identifier = ctx.port_identifier()
        implicit_data_type = ctx.implicit_data_type()
        data_type_or_implicit = ctx.data_type_or_implicit()
        data_type = ctx.data_type()
        constant_expression = ctx.constant_expression()

        if port_direction:
            parsed_port_direction = port_direction.getText()
            parsed_port_direction = (
                parsed_port_direction if parsed_port_direction != "ref" else "input"
            )
            if port_identifier:
                parsed_port_identifier = port_identifier.getText()
                port_dimension = "scalar"
                if implicit_data_type:
                    port_dimension = self.visitImplicit_data_type(implicit_data_type)
                elif data_type_or_implicit:
                    port_dimension = self.visitData_type_or_implicit(
                        data_type_or_implicit
                    )
                elif data_type:
                    port_dimension = self.visitData_type(data_type)

                return Port(
                    name=parsed_port_identifier,
                    direction=parsed_port_direction,
                    dimension=port_dimension,
                    has_default=constant_expression is not None,
                )
        return None

    def visitParameter_port_list(
        self, ctx: SystemVerilogParser.Parameter_port_listContext
    ) -> list[AntlrParameter]:
        parameters: list[AntlrParameter]

        list_of_param_assignments = ctx.list_of_param_assignments()
        parameter_port_declarations = ctx.parameter_port_declaration()

        parameters = []
        if list_of_param_assignments:
            parameters = self.visitList_of_param_assignments(list_of_param_assignments)
        if parameter_port_declarations:
            for parameter_port_declaration in parameter_port_declarations:
                parameters += self.visitParameter_port_declaration(
                    parameter_port_declaration
                )

        return parameters

    def visitParameter_port_declaration(
        self, ctx: SystemVerilogParser.Parameter_port_declarationContext
    ) -> list[AntlrParameter]:

        parameter_declaration = ctx.parameter_declaration()
        local_parameter_declaration = ctx.local_parameter_declaration()
        list_of_param_assignments = ctx.list_of_param_assignments()

        if parameter_declaration:
            return self.visitParameter_declaration(parameter_declaration)
        if local_parameter_declaration:
            return self.visitLocal_parameter_declaration(local_parameter_declaration)
        if list_of_param_assignments:
            return self.visitList_of_param_assignments(list_of_param_assignments)
        return []

    def visitModule_identifier(
        self, ctx: SystemVerilogParser.Module_identifierContext
    ) -> str:
        identifier = ctx.identifier()
        if identifier:
            return identifier.getText()

        raise IllegalStateException("Found none identifier in module identifier")

    def visitPackage_import_declaration(
        self, ctx: SystemVerilogParser.Package_import_declarationContext
    ) -> list[tuple[str, str]]:
        parsed_package_import_items: list[tuple[str, str]]

        package_import_items = ctx.package_import_item()
        if package_import_items:
            parsed_package_import_items = []
            for package_import_item in package_import_items:
                parsed_package_import_items.append(
                    self.visitPackage_import_item(package_import_item)
                )
            return parsed_package_import_items
        else:
            raise IllegalStateException(
                "An import declaration must have at least one item"
            )

    def visitPackage_import_item(
        self, ctx: SystemVerilogParser.Package_import_itemContext
    ) -> tuple[str, str]:
        parsed_identifier: str
        parsed_package_identifier: str

        package_identifier = ctx.package_identifier()
        identifier = ctx.identifier()

        if package_identifier:
            parsed_package_identifier = self.visitPackage_identifier(package_identifier)
        else:
            raise IllegalStateException(
                "A package import must have exactly one identifier"
            )

        if identifier:
            parsed_identifier = identifier.getText()
        else:
            parsed_identifier = "*"

        return parsed_package_identifier, parsed_identifier

    def visitPackage_identifier(
        self, ctx: SystemVerilogParser.Package_identifierContext
    ) -> str:
        identifier = ctx.identifier()
        if identifier:
            return identifier.getText()
        raise IllegalStateException(
            "A package identifier must contain exactly one identifier"
        )

    def visitModule_item(
        self, ctx: SystemVerilogParser.Module_itemContext
    ) -> Optional[Union[list[Port], list[AntlrParameter], HdlAntlrModule]]:
        # parameter override is not supported
        # neither with the "#-syntax" nor with the defparam syntax
        # program blocks are ignored
        # attributes are ignored
        # timeunits declarations are ignored

        port_declaration = ctx.port_declaration()
        module_common_item = ctx.module_common_item()
        specparam_declaration = ctx.specparam_declaration()
        specify_block = ctx.specify_block()
        module_declaration = ctx.module_declaration()
        interface_declaration = ctx.interface_declaration()

        if port_declaration:
            # list[Port]
            return self.visitPort_declaration(port_declaration)
        elif module_common_item:
            # list[AntlrParameter]
            return self.visitModule_common_item(module_common_item)
        elif specparam_declaration:
            # list[AntlrParameter]
            return self.visitSpecparam_declaration(specparam_declaration)
        elif specify_block:
            # list[AntlrParameter]
            return self.visitSpecify_block(specify_block)
        elif module_declaration:
            # HdlAntlrModule
            return self.visitModule_declaration(module_declaration)
        elif interface_declaration:
            # HdlAntlrModule
            return self.visitInterface_declaration(interface_declaration)
        return None

    def visitPort_declaration(
        self, ctx: SystemVerilogParser.Port_declarationContext
    ) -> list[Port]:
        # interface ports are ignored

        inout_declaration = ctx.inout_declaration()
        input_declaration = ctx.input_declaration()
        output_declaration = ctx.output_declaration()
        ref_declaration = ctx.ref_declaration()

        if inout_declaration:
            return self.visitInout_declaration(inout_declaration)

        if input_declaration:
            return self.visitInput_declaration(input_declaration)

        if output_declaration:
            return self.visitOutput_declaration(output_declaration)

        if ref_declaration:
            return self.visitRef_declaration(ref_declaration)
        return []

    def visitInterface_declaration(
        self, ctx: SystemVerilogParser.Interface_declarationContext
    ) -> HdlAntlrModule:
        def parse_interface_items(
            interface_items: list[SystemVerilogParser.Interface_itemContext],
        ) -> tuple[list[Port], list[AntlrParameter], list[HdlAntlrModule]]:
            ports = []
            parameters = []
            submodules = []
            for interface_item in interface_items:
                parsed_interface_item = self.visitInterface_item(interface_item)
                if parsed_interface_item:
                    if isinstance(parsed_interface_item, list):
                        if isinstance(parsed_interface_item[0], Port):
                            parsed_interface_item = cast(
                                list[Port], parsed_interface_item
                            )
                            ports += parsed_interface_item
                        else:
                            parsed_interface_item = cast(
                                list[AntlrParameter], parsed_interface_item
                            )
                            parameters += parsed_interface_item
                    else:
                        submodules.append(parsed_interface_item)
                return ports, parameters, submodules
            return [], [], []

        module: HdlAntlrModule
        parsed_interface_identifier: str
        module: HdlAntlrModule
        ports: list[Port]
        parameters: list[AntlrParameter]
        submodules: list[HdlAntlrModule]

        interface_header = ctx.interface_header()
        interface_identifier = ctx.interface_identifier()
        interface_items = ctx.interface_item()

        ports = []
        parameters = []
        submodules = []
        if interface_header:
            module: HdlAntlrModule = self.visitInterface_header(interface_header)
            if interface_items:
                ports, parameters, submodules = parse_interface_items(interface_items)
            return HdlAntlrModule(
                name=module.name,
                parameters=module.parameters + parameters,
                ports=module.ports + ports,
                submodules=submodules,
            )

        elif interface_identifier:
            parsed_interface_identifier = self.visitInterface_identifier(
                interface_identifier
            )
            if interface_items:
                ports, parameters, submodules = parse_interface_items(interface_items)
            return HdlAntlrModule(
                name=parsed_interface_identifier,
                ports=ports,
                parameters=parameters,
                submodules=submodules,
            )

        raise IllegalStateException("Empty interface declaration is forbidden")

    def visitInterface_item(
        self, ctx: SystemVerilogParser.Interface_itemContext
    ) -> Optional[Union[list[Port], list[AntlrParameter], HdlAntlrModule]]:

        port_declaration = ctx.port_declaration()
        module_common_item = ctx.module_common_item()
        interface_declaration = ctx.interface_declaration()

        if port_declaration:
            # list[Port]
            return self.visitPort_declaration(port_declaration)
        if module_common_item:
            # list[AntlrParameter]
            return self.visitModule_common_item(module_common_item)
        if interface_declaration:
            # HdlAntlrModule
            return self.visitInterface_declaration(interface_declaration)
        return None

    def visitInterface_header(
        self, ctx: SystemVerilogParser.Interface_headerContext
    ) -> HdlAntlrModule:
        interface_name: str
        parameters: list[AntlrParameter]
        ports: list[Port]
        interface_identifier = ctx.interface_identifier()
        parameter_port_list = ctx.parameter_port_list()
        list_of_port_declarations = ctx.list_of_port_declarations()

        if interface_identifier:
            interface_name = self.visitInterface_identifier(interface_identifier)

            parameters = []
            if parameter_port_list:
                parameters = self.visitParameter_port_list(parameter_port_list)

            ports = []
            if list_of_port_declarations:
                ports = self.visitList_of_port_declarations(list_of_port_declarations)

            return HdlAntlrModule(
                name=interface_name, parameters=parameters, ports=ports
            )
        raise IllegalStateException("An interface header must hold an identifier")

    def visitInterface_identifier(
        self, ctx: SystemVerilogParser.Interface_identifierContext
    ) -> str:
        identifier = ctx.identifier
        if identifier:
            return identifier.getText()

        raise IllegalStateException("Interface identifier must hold an identifier")

    def visitRef_declaration(
        self, ctx: SystemVerilogParser.Ref_declarationContext
    ) -> list[Port]:
        variable_identifiers: list[str]
        port_dimension: PORT_DIMENSION

        list_of_variable_identifiers = ctx.list_of_variable_identifiers()
        variable_port_type = ctx.variable_port_type()

        if list_of_variable_identifiers:
            variable_identifiers = self.visitList_of_variable_identifiers(
                list_of_variable_identifiers
            )

            if variable_port_type:
                port_dimension = self.visitVariable_port_type(variable_port_type)
            else:
                raise IllegalStateException(
                    "Variable port type is mandatory for ref declaration"
                )

            return [
                Port(
                    name=variable_identifier,
                    direction="input",
                    dimension=port_dimension,
                    has_default=False,
                )
                for variable_identifier in variable_identifiers
            ]
        return []

    def visitVariable_port_type(
        self, ctx: SystemVerilogParser.Variable_port_typeContext
    ) -> PORT_DIMENSION:
        var_data_type = ctx.var_data_type()
        if var_data_type:
            return self.visitVar_data_type(var_data_type)
        raise IllegalStateException("Variable port type must hold a vard data type")

    def visitVar_data_type(self, ctx: SystemVerilogParser.Var_data_typeContext):
        data_type = ctx.data_type()
        data_type_or_implicit = ctx.data_type_or_implicit()

        if data_type:
            return self.visitData_type(data_type)
        if data_type_or_implicit:
            return self.visitData_type_or_implicit(data_type_or_implicit)

        return "scalar"

    def visitInout_declaration(
        self, ctx: SystemVerilogParser.Inout_declarationContext
    ) -> list[Port]:

        parsed_identifiers: list[str]
        parsed_port_size: PORT_DIMENSION
        ports: list[Port]

        list_of_port_identifiers = ctx.list_of_port_identifiers()
        net_port_type = ctx.net_port_type()

        if list_of_port_identifiers:
            parsed_identifiers = self.visitList_of_port_identifiers(
                list_of_port_identifiers
            )

            if net_port_type:
                parsed_port_size = self.visitNet_port_type(net_port_type)
            else:
                parsed_port_size = "scalar"

            ports = []
            for parsed_identifier in parsed_identifiers:
                ports.append(
                    Port(
                        name=parsed_identifier,
                        direction="inout",
                        dimension=parsed_port_size,
                        has_default=False,
                    )
                )

        return []

    def visitOutput_declaration(
        self, ctx: SystemVerilogParser.Input_declarationContext
    ) -> list[Port]:
        port_identifiers: list[str]
        variable_identifiers: list[str]
        port_dimension: PORT_DIMENSION

        list_of_port_identifiers = ctx.list_of_port_identifiers()
        list_of_variable_identifiers = ctx.list_of_variable_identifiers()
        implicit_data_type = ctx.implicit_data_type()
        data_type_or_implicit = ctx.data_type_or_implicit()
        data_type = ctx.data_type()

        if list_of_port_identifiers:
            port_identifiers = self.visitList_of_port_identifiers(
                list_of_port_identifiers
            )
            port_dimension = "scalar"
            if implicit_data_type:
                port_dimension = self.visitImplicit_data_type(implicit_data_type)
            elif data_type_or_implicit:
                port_dimension = self.visitData_type_or_implicit(data_type_or_implicit)
            return [
                Port(
                    name=port_identifier,
                    direction="output",
                    dimension=port_dimension,
                    has_default=False,
                )
                for port_identifier in port_identifiers
            ]
        elif list_of_variable_identifiers:
            variable_identifiers = self.visitList_of_variable_identifiers(
                list_of_variable_identifiers
            )
            port_dimension = "scalar"
            if data_type_or_implicit:
                port_dimension = self.visitData_type_or_implicit(data_type_or_implicit)
            if data_type:
                port_dimension = self.visitData_type(data_type)
            return [
                Port(
                    name=port_identifier,
                    direction="output",
                    dimension=port_dimension,
                    has_default=False,
                )
                for port_identifier in variable_identifiers
            ]

        return []

    def visitInput_declaration(
        self, ctx: SystemVerilogParser.Input_declarationContext
    ) -> list[Port]:
        port_identifiers: list[str]
        variable_identifiers: list[str]
        port_dimension: PORT_DIMENSION

        list_of_port_identifiers = ctx.list_of_port_identifiers()
        list_of_variable_identifiers = ctx.list_of_variable_identifiers()
        implicit_data_type = ctx.implicit_data_type()
        data_type_or_implicit = ctx.data_type_or_implicit()
        data_type = ctx.data_type()

        if list_of_port_identifiers:
            port_identifiers = self.visitList_of_port_identifiers(
                list_of_port_identifiers
            )
            port_dimension = "scalar"
            if implicit_data_type:
                port_dimension = self.visitImplicit_data_type(implicit_data_type)
            elif data_type_or_implicit:
                port_dimension = self.visitData_type_or_implicit(data_type_or_implicit)
            return [
                Port(
                    name=port_identifier,
                    direction="input",
                    dimension=port_dimension,
                    has_default=False,
                )
                for port_identifier in port_identifiers
            ]
        elif list_of_variable_identifiers:
            variable_identifiers = self.visitList_of_variable_identifiers(
                list_of_variable_identifiers
            )
            port_dimension = "scalar"
            if data_type_or_implicit:
                port_dimension = self.visitData_type_or_implicit(data_type_or_implicit)
            if data_type:
                port_dimension = self.visitData_type(data_type)
            return [
                Port(
                    name=port_identifier,
                    direction="input",
                    dimension=port_dimension,
                    has_default=False,
                )
                for port_identifier in variable_identifiers
            ]

        return []

    def visitList_of_variable_identifiers(
        self, ctx: SystemVerilogParser.List_of_variable_identifiersContext
    ) -> list[str]:
        variables: list[str]

        var_ids = ctx.var_id()
        if var_ids:
            variables = []
            for var_id in var_ids:
                variables.append(self.visitVar_id(var_id))
            return variables
        raise IllegalStateException(
            "A list of variable identifiers must contain at least one identifier"
        )

    def visitVar_id(self, ctx: SystemVerilogParser.Var_idContext) -> str:

        variable_identifier = ctx.variable_identifier()

        if variable_identifier:
            return variable_identifier.getText()
        raise IllegalStateException("A var_id must contain a variable identifier")

    def visitNet_port_type(
        self, ctx: SystemVerilogParser.Net_port_typeContext
    ) -> PORT_DIMENSION:
        data_type_or_implicit = ctx.data_type_or_implicit()
        implicit_data_type = ctx.implicit_data_type()

        if data_type_or_implicit:
            return self.visitData_type_or_implicit(data_type_or_implicit)
        if implicit_data_type:
            return self.visitImplicit_data_type(implicit_data_type)

        return "scalar"

    def visitData_type_or_implicit(
        self, ctx: SystemVerilogParser.Data_type_or_implicitContext
    ) -> PORT_DIMENSION:
        data_type = ctx.data_type()
        implicit_data_type = ctx.implicit_data_type()

        if data_type:
            return self.visitData_type(data_type)
        if implicit_data_type:
            return self.visitImplicit_data_type(implicit_data_type)

        raise IllegalStateException(
            "Data type or implicit needs either data type or implicit"
        )

    def visitData_type(
        self, ctx: SystemVerilogParser.Data_typeContext
    ) -> PORT_DIMENSION:
        packed_dimension = ctx.packed_dimension()
        if packed_dimension:
            return "vectorial"
        else:
            return "scalar"

    def visitImplicit_data_type(
        self, ctx: SystemVerilogParser.Implicit_data_typeContext
    ) -> PORT_DIMENSION:
        packed_dimension = ctx.packed_dimension()

        if packed_dimension:
            return "vectorial"
        else:
            return "scalar"

    def visitList_of_port_identifiers(
        self, ctx: SystemVerilogParser.List_of_port_identifiersContext
    ) -> list[str]:
        parsed_port_ids: list[str]

        port_ids = ctx.port_id()

        if port_ids:
            parsed_port_ids = []
            for port_id in port_ids:
                parsed_port_ids.append(port_id.getText())
            return parsed_port_ids
        return []

    def visitSpecify_block(
        self, ctx: SystemVerilogParser.Specify_blockContext
    ) -> list[AntlrParameter]:
        parsed_specify_items: list[AntlrParameter]

        specify_items = ctx.specify_item()
        if specify_items:
            parsed_specify_items = []
            for specify_item in specify_items:
                parameters = self.visitSpecify_item(specify_item)
                if parameters:
                    parsed_specify_items += parameters
            return parsed_specify_items

        return []

    def visitSpecify_item(
        self, ctx: SystemVerilogParser.Specify_itemContext
    ) -> list[AntlrParameter]:
        specparam_declaration = ctx.specparam_declaration()

        if specparam_declaration:
            return self.visitSpecparam_declaration(specparam_declaration)

        return []

    def visitSpecparam_declaration(
        self, ctx: SystemVerilogParser.Specparam_declarationContext
    ) -> list[AntlrParameter]:
        list_of_specparam_assignments = ctx.list_of_specparam_assignments()

        if list_of_specparam_assignments:
            return self.visitList_of_specparam_assignments(
                list_of_specparam_assignments
            )

        return []

    def visitList_of_specparam_assignments(
        self, ctx: SystemVerilogParser.List_of_specparam_assignmentsContext
    ) -> list[AntlrParameter]:
        parsed_spec_param_assignments: list[AntlrParameter]

        spec_param_assignments = ctx.specparam_assignment()
        if spec_param_assignments:
            parsed_spec_param_assignments = []
            for spec_param_assignment in spec_param_assignments:
                parameter = self.visitSpecparam_assignment(spec_param_assignment)
                if parameter:
                    parsed_spec_param_assignments.append(parameter)
            return parsed_spec_param_assignments

        raise IllegalStateException(
            "Specparam assignment list must have at least one item"
        )

    def visitSpecparam_assignment(
        self, ctx: SystemVerilogParser.Specparam_assignmentContext
    ) -> Optional[AntlrParameter]:
        # pulse control is ignored

        parsed_identifier: str
        specparam_identifier = ctx.specparam_identifier()
        expression = ctx.constant_mintypmax_expression()

        if specparam_identifier:
            parsed_identifier = specparam_identifier.getText()
            if expression:
                return AntlrParameter(
                    name=parsed_identifier,
                    value=expression.getText(),
                    rule=expression,
                    has_default=True,
                )
            else:
                raise IllegalStateException("Specparams must have an assigned values")

        return None

    def visitModule_common_item(
        self, ctx: SystemVerilogParser.Module_common_itemContext
    ) -> list[AntlrParameter]:
        module_item_declaration = ctx.module_item_declaration()

        if module_item_declaration:
            return self.visitModule_item_declaration(module_item_declaration)

        return []

    def visitModule_item_declaration(
        self, ctx: SystemVerilogParser.Module_item_declarationContext
    ) -> list[AntlrParameter]:
        # genvar and clocking declarations are ignored

        package_item_declaration = ctx.package_item_declaration()

        if package_item_declaration:
            return self.visitPackage_item_declaration(package_item_declaration)

        return []

    def visitPackage_item_declaration(
        self, ctx: SystemVerilogParser.Package_item_declarationContext
    ) -> list[AntlrParameter]:
        # we only parse local_parameters and parameters
        # all other constructs (classes, interface, functions...)
        # are ignored

        local_parameter_declaration = ctx.local_parameter_declaration()
        parameter_declaration = ctx.parameter_declaration()

        if local_parameter_declaration:
            return self.visitLocal_parameter_declaration(local_parameter_declaration)
        elif parameter_declaration:
            return self.visitParameter_declaration(parameter_declaration)

        return []

    def visitLocal_parameter_declaration(
        self, ctx: SystemVerilogParser.Local_parameter_declarationContext
    ) -> list[AntlrParameter]:
        # localparam types are ignored
        # localparam type handle definitions are ignored

        list_of_param_assignments = ctx.list_of_param_assignments()
        if list_of_param_assignments:
            return self.visitList_of_param_assignments(list_of_param_assignments)

        return []

    def visitParameter_declaration(
        self, ctx: SystemVerilogParser.Local_parameter_declarationContext
    ) -> list[AntlrParameter]:
        # param types are ignored
        # param type handle definitions are ignored

        list_of_param_assignments = ctx.list_of_param_assignments()
        if list_of_param_assignments:
            return self.visitList_of_param_assignments(list_of_param_assignments)

        return []

    def visitList_of_param_assignments(
        self, ctx: SystemVerilogParser.List_of_param_assignmentsContext
    ) -> list[AntlrParameter]:
        parsed_param_assignments: list[AntlrParameter]

        param_assignments = ctx.param_assignment()
        if param_assignments:
            parsed_param_assignments = []
            for param_assignment in param_assignments:
                parsed_param_assignments.append(
                    self.visitParam_assignment(param_assignment)
                )
            return parsed_param_assignments

        raise IllegalStateException(
            "A list of parameter assignments must include at least one parameter"
        )

    def visitParam_assignment(
        self, ctx: SystemVerilogParser.Param_assignmentContext
    ) -> AntlrParameter:
        # eventual dimensions are skipped
        parsed_parameter_identifier: str
        parsed_parameter_value: str
        parameter_value_context: SystemVerilogParser.Constant_param_expressionContext
        has_default: bool

        parameter_identifier = ctx.parameter_identifier()
        constant_param_expression = ctx.constant_param_expression()

        if parameter_identifier:
            parsed_parameter_identifier = parameter_identifier.getText()

            if constant_param_expression:
                parsed_parameter_value = constant_param_expression.getText()
                parameter_value_context = constant_param_expression
                has_default = True
            else:
                parsed_parameter_value = ""
                parameter_value_context = parameter_identifier
                has_default = False

            return AntlrParameter(
                name=parsed_parameter_identifier,
                value=parsed_parameter_value,
                rule=parameter_value_context,
                has_default=has_default,
            )
        raise IllegalStateException(
            "Parameter identifier is mandatory in parameter assignment"
        )

    @staticmethod
    def multiple_not_none_items(l: list[Any]):
        return any(
            [
                (True if context_1 is not None and context_2 is not None else False)
                for context_1, context_2 in zip(l, l)
            ]
        )
